@@ -24,6 +24,7 @@
 
 #include "config.h"
 
+#define _GNU_SOURCE
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
@@ -43,6 +44,7 @@
 
 #include "file-transfer-dialog.h"
 
+#include "gs-theme-manager.h"
 #include "gs-job.h"
 #include "gs-prefs.h" /* for GS_MODE enum */
 
@@ -82,8 +84,9 @@ static GtkTargetEntry drop_types [] =
         { "_NETSCAPE_URL", 0, TARGET_NS_URL }
 };
 
-static GladeXML *xml = NULL;
-static GSJob    *job = NULL;
+static GladeXML       *xml = NULL;
+static GSThemeManager *theme_manager = NULL;
+static GSJob          *job = NULL;
 
 static gint32
 config_get_activate_delay (gboolean *is_writable)
@@ -214,18 +217,18 @@ config_get_theme (gboolean *is_writable)
 }
 
 static GSList *
-get_all_theme_ids (GSJob *job)
+get_all_theme_ids (GSThemeManager *theme_manager)
 {
         GSList *ids = NULL;
         GSList *entries;
         GSList *l;
 
-        entries = gs_job_get_theme_info_list (job);
+        entries = gs_theme_manager_get_info_list (theme_manager);
         for (l = entries; l; l = l->next) {
-                GSJobThemeInfo *info = l->data;
+                GSThemeInfo *info = l->data;
 
-                ids = g_slist_prepend (ids, g_strdup (gs_job_theme_info_get_id (info)));
-                gs_job_theme_info_unref (info);
+                ids = g_slist_prepend (ids, g_strdup (gs_theme_info_get_id (info)));
+                gs_theme_info_unref (info);
         }
         g_slist_free (entries);
 
@@ -247,7 +250,7 @@ config_set_theme (const char *theme_id)
                 mode = GS_MODE_RANDOM;
 
                 /* set the themes key to contain all available screensavers */
-                list = get_all_theme_ids (job);
+                list = get_all_theme_ids (theme_manager);
         } else {
                 mode = GS_MODE_SINGLE;
                 list = g_slist_append (list, g_strdup (theme_id));
@@ -342,16 +345,36 @@ preview_clear (GtkWidget *widget)
         gdk_window_clear (widget->window);
 }
 
+static void
+job_set_theme (GSJob      *job,
+               const char *theme)
+{
+        GSThemeInfo *info;
+        const char  *command;
+
+        command = NULL;
+
+        info = gs_theme_manager_lookup_theme_info (theme_manager, theme);
+        if (info != NULL) {
+                command = gs_theme_info_get_exec (info);
+        }
+
+        gs_job_set_command (job, command);
+
+        if (info != NULL) {
+                gs_theme_info_unref (info);
+        }
+}
+
 static void
 preview_set_theme (GtkWidget  *widget,
                    const char *theme,
                    const char *name)
 {
-        GError    *error = NULL;
         GtkWidget *label;
         char      *markup;
 
-        if (job) {
+        if (job != NULL) {
                 gs_job_stop (job);
         }
 
@@ -367,28 +390,16 @@ preview_set_theme (GtkWidget  *widget,
         } else if (theme && strcmp (theme, "__random") == 0) {
                 GSList *themes;
 
-                themes = get_all_theme_ids (job);
-                if (themes) {
-                        if (! gs_job_set_theme (job, (const char *) themes->data, &error)) {
-                                if (error) {
-                                        g_warning ("Could not set theme: %s", error->message);
-                                        g_error_free (error);
-                                }
-                                return;
-                        }
+                themes = get_all_theme_ids (theme_manager);
+                if (themes != NULL) {
+                        job_set_theme (job, (const char *) themes->data);
                         g_slist_foreach (themes, (GFunc) g_free, NULL);
                         g_slist_free (themes);
 
                         gs_job_start (job);
                 }
         } else {
-                if (! gs_job_set_theme (job, theme, &error)) {
-                        if (error) {
-                                g_warning ("Could not set theme: %s", error->message);
-                                g_error_free (error);
-                        }
-                        return;
-                }
+                job_set_theme (job, theme);
                 gs_job_start (job);
         }
 }
@@ -421,7 +432,7 @@ response_cb (GtkWidget *widget,
 static GSList *
 get_theme_info_list (void)
 {
-        return gs_job_get_theme_info_list (job);
+        return gs_theme_manager_get_info_list (theme_manager);
 }
 
 static void
@@ -456,18 +467,18 @@ populate_model (GtkTreeStore *store)
         }
 
         for (l = themes; l; l = l->next) {
-                GSJobThemeInfo *info = l->data;
+                GSThemeInfo *info = l->data;
 
                 if (! info)
                         continue;
 
                 gtk_tree_store_append (store, &iter, NULL);
                 gtk_tree_store_set (store, &iter,
-                                    NAME_COLUMN, gs_job_theme_info_get_name (info),
-                                    ID_COLUMN, gs_job_theme_info_get_id (info),
+                                    NAME_COLUMN, gs_theme_info_get_name (info),
+                                    ID_COLUMN, gs_theme_info_get_id (info),
                                     -1);
 
-                gs_job_theme_info_unref (info);
+                gs_theme_info_unref (info);
         }
 
         g_slist_free (themes);
@@ -521,7 +532,7 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
 
         gtk_tree_model_get (model, &iter, ID_COLUMN, &theme, NAME_COLUMN, &name, -1);
 
-        if (! theme) {
+        if (theme == NULL) {
                 g_free (name);
                 return;
         }
@@ -763,26 +774,38 @@ theme_installer_run (GtkWidget *parent,
 {
         GtkWidget    *dialog;
         GnomeVFSURI  *src_uri;
-        GList        *src, *target;
-        char         *target_path;
+        GList        *src;
+        GList        *target;
         char         *user_dir;
-        char         *base;
+        char         *short_name;
+        char         *base_name;
+        char         *target_path;
+        gboolean      is_desktop;
 
         src_uri = gnome_vfs_uri_new (filename);
         src = g_list_append (NULL, src_uri);
-
-        user_dir = g_build_filename (g_get_user_data_dir (), "gnome-screensaver", "themes", NULL);
-        base = gnome_vfs_uri_extract_short_name (src_uri);
-
+        target = NULL;
         target_path = NULL;
 
-        while (TRUE) {
-                char      *file_tmp;
-                int        len = strlen (base);
+        user_dir = g_build_filename (g_get_user_data_dir (), "applications", "screensavers", NULL);
+        g_mkdir_with_parents (user_dir, S_IRWXU);
 
-                if (base && len > 4 && (!strcmp (base + len - 4, ".xml"))) {
-                        file_tmp = g_strdup_printf ("screensaver-theme-%d.xml", rand ());
-                } else {
+        short_name = gnome_vfs_uri_extract_short_name (src_uri);
+        base_name = NULL;
+
+        is_desktop = FALSE;
+        if (short_name != NULL
+            && g_str_has_suffix (short_name, ".desktop")) {
+                /* FIXME: validate key file? */
+                is_desktop = TRUE;
+                base_name = g_strndup (short_name, strlen (short_name) - 8);
+        }
+
+        while (TRUE) {
+                char *tmp;
+                gboolean exists;
+
+                if (! is_desktop) {
                         dialog = gtk_message_dialog_new (GTK_WINDOW (parent),
                                                          GTK_DIALOG_MODAL,
                                                          GTK_MESSAGE_ERROR,
@@ -797,20 +820,25 @@ theme_installer_run (GtkWidget *parent,
                         gtk_dialog_run (GTK_DIALOG (dialog));
                         gtk_widget_destroy (dialog);
                         g_free (target_path);
-                        return;
+                        goto out;
                 }
 
-                target_path = g_build_filename (user_dir, file_tmp, NULL);
+                g_free (target_path);
+                target_path = g_build_filename (user_dir, short_name, NULL);
 
-                g_free (file_tmp);
-                if (! gnome_vfs_uri_exists (gnome_vfs_uri_new (target_path)))
+                /* FIXME: racy, oh well */
+                exists = gnome_vfs_uri_exists (gnome_vfs_uri_new (target_path));
+                if (! exists) {
+                        target = g_list_append (NULL, gnome_vfs_uri_new (target_path));
                         break;
+                }
+
+                /* try another name */
+                tmp = g_strdup_printf ("%s-%u.desktop",  base_name, g_random_int ());
+                g_free (short_name);
+                short_name = tmp;
         }
 
-        g_free (user_dir);
-
-        target = g_list_append (NULL, gnome_vfs_uri_new (target_path));
-
         dialog = file_transfer_dialog_new ();
         gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (parent));
         gtk_window_set_icon_name (GTK_WINDOW (dialog), "screensaver");
@@ -821,8 +849,6 @@ theme_installer_run (GtkWidget *parent,
                                               GNOME_VFS_XFER_ERROR_MODE_QUERY,
                                               GNOME_VFS_XFER_OVERWRITE_MODE_QUERY,
                                               GNOME_VFS_PRIORITY_DEFAULT);
-        gnome_vfs_uri_list_unref (src);
-        gnome_vfs_uri_list_unref (target);
 
         g_signal_connect (dialog, "cancel",
                           G_CALLBACK (transfer_cancel_cb), target_path);
@@ -830,6 +856,14 @@ theme_installer_run (GtkWidget *parent,
                           G_CALLBACK (transfer_done_cb), target_path);
 
         gtk_widget_show (dialog);
+
+ out:
+        gnome_vfs_uri_list_unref (src);
+        gnome_vfs_uri_list_unref (target);
+
+        g_free (user_dir);
+        g_free (base_name);
+        g_free (short_name);
 }
 
 /* Callback issued during drag movements */
@@ -876,14 +910,13 @@ drag_data_received_cb (GtkWidget        *widget,
         if (uris != NULL && uris->data != NULL) {
                 GnomeVFSURI *uri = (GnomeVFSURI *) uris->data;
 
-                if (gnome_vfs_uri_is_local (uri))
-                        filename = gnome_vfs_unescape_string (
-                                        gnome_vfs_uri_get_path (uri),
-                                        G_DIR_SEPARATOR_S);
-                else
-                        filename = gnome_vfs_unescape_string (
-                                        gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE),
-                                        G_DIR_SEPARATOR_S);
+                if (gnome_vfs_uri_is_local (uri)) {
+                        filename = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri),
+                                                              G_DIR_SEPARATOR_S);
+                } else {
+                        filename = gnome_vfs_unescape_string (gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE),
+                                                              G_DIR_SEPARATOR_S);
+                }
 
                 gnome_vfs_uri_list_unref (uris);
         }
@@ -1183,11 +1216,6 @@ check_is_root_user (void)
   gid_t rgid, egid, sgid; /* Real, effective and saved group ID's */
 
 #ifdef HAVE_GETRESUID
-  /* These aren't in the header files, so we prototype them here.
-   */
-  int getresuid(uid_t *ruid, uid_t *euid, uid_t *suid);
-  int getresgid(gid_t *rgid, gid_t *egid, gid_t *sgid);
-
   if (getresuid (&ruid, &euid, &suid) != 0 ||
       getresgid (&rgid, &egid, &sgid) != 0)
 #endif /* HAVE_GETRESUID */
@@ -1429,7 +1457,7 @@ init_capplet (void)
                 g_free (string);
 
                 if (mode == GS_MODE_RANDOM) {
-                        list = get_all_theme_ids (job);
+                        list = get_all_theme_ids (theme_manager);
                         gconf_client_set_list (client, KEY_THEMES, GCONF_VALUE_STRING, list, NULL);
 
                         g_slist_foreach (list, (GFunc) g_free, NULL);
@@ -1494,11 +1522,13 @@ main (int    argc,
                             NULL);
 
         job = gs_job_new ();
+        theme_manager = gs_theme_manager_new ();
 
         init_capplet ();
 
         gtk_main ();
 
+        g_object_unref (theme_manager);
         g_object_unref (job);
 
 	return 0;
