@@ -1337,6 +1337,8 @@ gs_manager_create_window (GSManager *manager,
 
         n_monitors = gdk_screen_get_n_monitors (screen);
 
+        gs_debug ("Creating %d windows for screen %d", n_monitors, gdk_screen_get_number (screen));
+
         for (i = 0; i < n_monitors; i++) {
                 window = gs_window_new (screen, i, manager->priv->lock_active);
 
@@ -1430,7 +1432,7 @@ gs_manager_activate (GSManager *manager)
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
         if (manager->priv->active) {
-                g_warning ("Trying to activate manager when already active");
+                gs_debug ("Trying to activate manager when already active");
                 return FALSE;
         }
 
@@ -1479,7 +1481,7 @@ gs_manager_deactivate (GSManager *manager)
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
         if (! manager->priv->active) {
-                g_warning ("Trying to deactivate a screensaver that is not active");
+                gs_debug ("Trying to deactivate a screensaver that is not active");
                 return FALSE;
         }
 
